import argparse
import logging
import os
import yaml

from collections import OrderedDict
from importlib import import_module

from .command import Command, AbortException

yaml.warnings({'YAMLLoadWarning': False})

SCRIPT_FIELDS = OrderedDict()
SCRIPT_FIELDS['systems'] = 'Systems'
SCRIPT_FIELDS['vars'] = 'Variables'


def read_yaml(yamlfile: str) -> dict:
    with open(yamlfile) as file:
        return yaml.load(file.read())


CONFIG = {
    'script_dir': '~/automatix-config',
    'encoding': os.getenv('ENCODING', 'utf-8'),
    'import_path': '.',
    'ssh_cmd': 'ssh {hostname} sudo ',
    'remote_tmp_dir': 'automatix_tmp',
    'logger': 'automatix',
    'bundlewrap': False,
    'teamvault': False,
}

configfile = os.getenv('AUTOMATIX_CONFIG', '~/.automatix.cfg.yaml')
if os.path.isfile(configfile):
    CONFIG.update(read_yaml(configfile))

if os.getenv('AUTOMATIX_SCRIPT_DIR'):
    CONFIG['script_dir'] = os.getenv('AUTOMATIX_SCRIPT_DIR')

if CONFIG.get('logging_lib'):
    log_lib = import_module(CONFIG.get('logging_lib'))
    init_logger = log_lib.init_logger
else:
    from .logger import init_logger

if CONFIG.get('bundlewrap'):
    from bundlewrap.repo import Repository
    CONFIG['bw_repo'] = Repository(repo_path=os.environ.get('BW_REPO_PATH'))

LOG = logging.getLogger(CONFIG['logger'])

SCRIPT_PATH = CONFIG['script_dir']

if CONFIG['teamvault']:
    import bwtv

    SCRIPT_FIELDS['secrets'] = 'Secrets'


    class UnknownSecretTypeException(Exception):
        pass


def _arguments():
    parser = argparse.ArgumentParser(
        description='Process automation tool',
    )
    parser.add_argument(
        'scriptfile',
        help='Path to scriptfile (yaml), use " -- " if needed to delimit this from argument fields',
    )
    for field in SCRIPT_FIELDS.keys():
        parser.add_argument(
            f'--{field}',
            nargs='*',
            help=f'Use this to set {field} without adding them to the script or to overwrite them. '
            f'You can specify multiple {field} like: --{field} v1=string1 v2=string2 v3=string3',
        )
    parser.add_argument(
        '--print-overview', '-p',
        action='store_true',
        help='Just print command pipeline overview with indices then exit without executing the commandline. '
             'Note that the *always pipeline* will be executed beforehand.',
    )
    parser.add_argument(
        '--jump-to', '-j',
        help='Jump to step instead of starting at the beginning',
        default=0,
    )
    parser.add_argument(
        '--interactive', '-i',
        action='store_true',
        help='Confirm actions before executing',
    )
    parser.add_argument(
        '--force', '-f',
        action='store_true',
        help='try always to proceed (except manual steps), even if errors occur (no retries)'
    )
    parser.add_argument(
        '--debug', '-d',
        action='store_true',
        help='activate debug log level',
    )
    return parser.parse_args()


def _overwrite(script: dict, key: str, data: str):
    script.setdefault(key, {})
    for item in data:
        k, v = item.split('=')
        script[key][k] = v


def get_script(args: argparse.Namespace) -> dict:
    file = args.scriptfile
    if not os.path.isfile(args.scriptfile):
        file = f'{SCRIPT_PATH}/{args.scriptfile}'

    script = read_yaml(file)

    for field in SCRIPT_FIELDS.keys():
        if vars(args).get(field):
            _overwrite(script=script, key=field, data=vars(args)[field])

    return script


def collect_vars(script: dict) -> dict:
    var_dict = script.get('vars', {})
    script['vars'] = var_dict  # just for the case it was empty
    if CONFIG['teamvault']:
        for key, secret in script.get('secrets', {}).items():
            sid, field = secret.split('_')
            if field == 'password':
                var_dict[key] = bwtv.password(sid)
            elif field == 'username':
                var_dict[key] = bwtv.username(sid)
            elif field == 'file':
                var_dict[key] = bwtv.file(sid)
            else:
                raise UnknownSecretTypeException(field)
    for syskey, system in script.get('systems', {}).items():
        var_dict[f'system_{syskey}'] = system
    return var_dict


def build_command_list(script: dict, variables: dict, pipeline: str) -> [Command]:
    command_list = []
    for index, cmd in enumerate(script[pipeline]):
        new_cmd = Command(
            config=CONFIG,
            pipeline_cmd=cmd,
            index=index,
            systems=script.get('systems', {}),
            variables=variables,
            imports=script.get('imports', []),
        )
        command_list.append(new_cmd)
        if new_cmd.assignment:
            variables[new_cmd.assignment_var] = f'{{{new_cmd.assignment_var}}}'
    return command_list


def print_main_data(script: dict):
    LOG.info(f"\nName: {script['name']}")
    for fieldkey, fieldvalue in SCRIPT_FIELDS.items():
        LOG.info(f'\n{fieldvalue}:')
        for key, value in script.get(fieldkey, {}).items():
            LOG.info(f" {key}: {value}")


def print_command_line_steps(command_list: [Command]):
    LOG.info('\nCommandline Steps:')
    for cmd in command_list:
        LOG.info(f"({cmd.index}) [{cmd.orig_key}]: {cmd.get_resolved_value()}")


def execute_pipeline(command_list: [Command], args: argparse.Namespace, start_index: int = 0):
    for cmd in command_list[start_index:]:
        cmd.execute(interactive=args.interactive, force=args.force)


def execute_extra_pipeline(script: dict, variables: dict, pipeline: str):
    try:
        if script.get(pipeline):
            pipeline_list = build_command_list(script=script, variables=variables, pipeline=pipeline)
            LOG.info('\n------------------------------')
            LOG.info(f' --- Start {pipeline.upper()} pipeline ---')
            execute_pipeline(command_list=pipeline_list, args=argparse.Namespace(interactive=False, force=False))
            LOG.info(f'\n --- End {pipeline.upper()} pipeline ---')
            LOG.info('------------------------------\n')
    except AbortException as exc:
        exit(int(str(exc)))
    except KeyboardInterrupt:
        LOG.warning('\nAborted by user. Exiting.')
        exit(130)


def main():
    args = _arguments()
    init_logger(name=CONFIG['logger'], debug=args.debug)

    script = get_script(args=args)

    for field in SCRIPT_FIELDS.keys():
        if vars(args).get(field):
            _overwrite(script=script, key=field, data=vars(args)[field])

    variables = collect_vars(script)

    command_list = build_command_list(script=script, variables=variables, pipeline='pipeline')

    execute_extra_pipeline(script=script, variables=variables, pipeline='always')

    print_main_data(script)
    print_command_line_steps(command_list)
    if args.print_overview:
        exit()

    try:
        execute_pipeline(command_list=command_list, args=args, start_index=int(args.jump_to))
    except AbortException as exc:
        LOG.debug('Abort requested. Cleaning up.')
        execute_extra_pipeline(script=script, variables=variables, pipeline='cleanup')
        LOG.debug('Clean up done. Exiting.')
        exit(int(str(exc)))
    except KeyboardInterrupt:
        LOG.warning('\nAborted by user. Exiting.')
        exit(130)

    execute_extra_pipeline(script=script, variables=variables, pipeline='cleanup')

    LOG.info('---------------------------------------------------------------')
    LOG.info('Automatix finished: Congratulations and have a N.I.C.E. day :-)')
    LOG.info('---------------------------------------------------------------')



CONSTANTS = {
    'apt_update':  'apt-get -qy update',
    'apt_upgrade': 'DEBIAN_FRONTEND=noninteractive apt-get -qy -o Dpkg::Options::=--force-confold --no-install-recommends upgrade',
    'apt_full_upgrade': 'DEBIAN_FRONTEND=noninteractive apt-get -qy -o Dpkg::Options::=--force-confold --no-install-recommends full-upgrade',
}



import logging
import os
import re
import subprocess

from shlex import quote

from .constants import CONSTANTS


class Command:
    def __init__(self, config: dict, pipeline_cmd: dict, index: int, systems: dict, variables: dict, imports: []):
        self.config = config
        self.pipeline_cmd = pipeline_cmd
        self.index = index
        self.systems = systems
        self.vars = variables
        self.imports = imports

        self.LOG = logging.getLogger(config['logger'])

        for key, value in pipeline_cmd.items():
            self.orig_key = key
            self.assignment, self.assignment_var, self.key = get_assignment_var(key=key)
            if isinstance(value, dict):
                # We need this workaround because the yaml lib returns a dictionary instead of a string,
                # if there is nothing but a variable in the command. Alternative is to use quotes in the script yaml.
                self.value = f'{{{next(iter(value))}}}'
            else:
                self.value = value
            break

    def get_type(self):
        if self.key == 'local':
            return 'local'
        if self.key == 'manual':
            return 'manual'
        if self.key == 'python':
            return 'python'
        if re.search('remote@', self.key):
            return 'remote'
        raise UnknownCommandException(f'Command type {self.key} is not known.')

    def get_system(self):
        if self.get_type() == 'remote':
            return self.systems[re.search('remote@(.*)', self.key).group(1)]
        return 'localhost'

    def get_resolved_value(self):
        variables = self.vars.copy()
        for key, value in CONSTANTS.items():
            variables[f'const_{key}'] = value
        return self.value.format(**variables)

    def execute(self, interactive: bool = False, force: bool = False):
        self.LOG.notice(f'\n({self.index}) [{self.orig_key}]: {self.get_resolved_value()}')
        return_code = 0

        if self.get_type() == 'manual' or interactive:
            answer = input(f'Proceed? (p: proceed (default), s: skip, a: abort)')
            if answer == 's':
                return
            if answer == 'a':
                raise AbortException(str(1))

        if self.get_type() == 'local':
            return_code = self._local_action()
        if self.get_type() == 'python':
            return_code = self._python_action()
        if self.get_type() == 'remote':
            return_code = self._remote_action()

        if return_code != 0:
            self.LOG.error(f'Command ({self.index}) failed with return code {return_code}.')
            if force:
                return
            err_answer = input('What should I do? (p: proceed (default), r: retry, a: abort)')
            if err_answer == 'r':
                return self.execute(interactive)
            if err_answer == 'a':
                raise AbortException(str(return_code))

    def _local_action(self) -> int:
        cmd = self._build_command(path=self.config['import_path'])
        try:
            return self._run_local_command(cmd=cmd)
        except KeyboardInterrupt:
            self.LOG.info('Abort command by user key stroke. Exit code is set to 130.')
            return 130

    def _python_action(self) -> int:
        cmd = self.get_resolved_value()
        if self.config['bundlewrap']:
            for key, value in self.systems.items():
                exec(f'{key}_node = self.config["bw_repo"].get_node(value)')
        try:
            self.LOG.debug(f'Run python command: {cmd}')
            if self.assignment_var:
                exec(f'self.vars[self.assignment_var] = {cmd}')
                self.LOG.info(f'Variable {self.assignment_var} = {self.vars[self.assignment_var]}')
            else:
                exec(cmd)
            return 0
        except KeyboardInterrupt:
            self.LOG.info('Abort command by user key stroke. Exit code is set to 130.')
            return 130
        except Exception as exc:
            self.LOG.error(exc)
            return 1

    def _remote_action(self) -> int:
        hostname = self.get_system()
        if self.config['bundlewrap']:
            node = self.config['bw_repo'].get_node(self.get_system())
            hostname = node.hostname

        ssh_cmd = self.config["ssh_cmd"].format(hostname=hostname)
        remote_cmd = self.get_resolved_value()
        prefix = ''
        if self.imports:
            prefix = f'tar -C {self.config["import_path"]} -cf - ' + ' '.join(self.imports) + ' | '
            remote_cmd = f'mkdir {self.config["remote_tmp_dir"]};' \
                f' tar -C {self.config["remote_tmp_dir"]} -xf -;' \
                f' {self._build_command(path=self.config["remote_tmp_dir"])}'

        cmd = f'{prefix}{ssh_cmd}{quote("bash -c " + quote(remote_cmd))}'

        try:
            exitcode = self._run_local_command(cmd=cmd)
        except KeyboardInterrupt:
            self.LOG.info('Abort command by user key stroke. Exit code is set to 130.')
            exitcode = 130

            ps_pids = self.get_remote_pids(hostname=hostname, cmd=self.get_resolved_value())
            while ps_pids:
                self.LOG.notice('Remote command seems still to be running! Found PIDs: {}'.format(','.join(ps_pids)))
                answer = input(
                    'What should I do? '
                    '(i: send SIGINT (default), t: send SIGTERM, k: send SIGKILL, p: do nothing and proceed) ')

                if answer == 'p':
                    break
                elif answer == 't':
                    signal = 'TERM'
                elif answer == 'k':
                    signal = 'KILL'
                else:
                    signal = 'INT'
                for pid in ps_pids:
                    kill_cmd = f'{ssh_cmd} kill -{signal} {pid}'
                    self.LOG.info(f'Kill {pid} on {hostname}')
                    subprocess.run(kill_cmd, shell=True)

                ps_pids = self.get_remote_pids(hostname=hostname, cmd=self.get_resolved_value())
            self.LOG.info('Keystroke interrupt handled.\n')

        if self.imports:
            cleanup_cmd = f'{ssh_cmd} rm -r {self.config["remote_tmp_dir"]}'
            self.LOG.debug(f'Executing: {cleanup_cmd}')
            proc = subprocess.run(cleanup_cmd, shell=True, executable='/bin/bash')
            if proc.returncode != 0:
                self.LOG.warning(f'Failed to remove {self.config["remote_tmp_dir"]}, exitcode: {proc.returncode}')

        return exitcode

    def _build_command(self, path: str) -> str:
        if not self.imports:
            return self.get_resolved_value()
        return f'. {path}/' + f'; . {path}/'.join(self.imports) + '; ' + self.get_resolved_value()

    def _run_local_command(self, cmd: str) -> int:
        self.LOG.debug(f'Executing: {cmd}')
        if self.assignment:
            proc = subprocess.run(cmd, shell=True, executable='/bin/bash', stdout=subprocess.PIPE)
            output = proc.stdout.decode(self.config["encoding"])
            self.vars[self.assignment_var] = output
            self.LOG.info(f'Variable {self.assignment_var} = {output}')
        else:
            proc = subprocess.run(cmd, shell=True, executable='/bin/bash')
        return proc.returncode

    def get_remote_pids(self, hostname, cmd) -> []:
        ps_cmd = f"ps axu | grep {quote(cmd)} | grep -v 'grep' | awk '{{print $2}}'"
        cmd = f'ssh {hostname} {quote(ps_cmd)} 2>&1'
        pids = subprocess.check_output(cmd, shell=True, executable='/bin/bash').decode(self.config["encoding"]).split()

        return pids


def get_assignment_var(key) -> (bool, str, str):
    if not re.search('=', key):
        return False, '', key
    return (True,) + re.search('(.*)=(.*)', key).group(1, 2)


class AbortException(Exception):
    pass


class UnknownCommandException(Exception):
    pass
